/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"
#include "CubeRegion.h"
#include "CallTree.h"
#include "TreeView.h"
#include "CubeMetric.h"
#include "AggregatedTreeItem.h"
#include "TreeModel.h"
#include "CubeProxy.h"
#include "CubeIdIndexMap.h"
#include "TabManager.h"

#include <QDesktopWidget>
#include <QLayout>
#include <QPropertyAnimation>

using namespace std;
using namespace cube;
using namespace cubegui;

CallTree::CallTree()
{
    treeType             = CALLTREE;
    iterationsAreHidden_ = false;
    loopRoot             = 0;
    aggregatedLoopRoot   = 0;
}

CallTree::~CallTree()
{
    unsetLoop(); // delete merged iterations
}

QString
CallTree::getLabel()
{
    if ( !cube )
    {
        return tr( "not yet initialized" );
    }
    return ( cube->getAttribute( "calltree title" ) != "" ) ? cube->getAttribute( "calltree title" ).c_str() : tr( "Call tree" );
}

TreeItem*
CallTree::createTree()
{
    TreeItem* top = new TreeItem( this, QString( "Invisible call tree root" ), CALLITEM, 0 );

    createItems<cube::Cnode>( top, cube->getRootCnodes(), CALLITEM );

    /** check if an item is defined as a loop ("CLUSTER ROOT CNODE ID" is set) */
    QString nrs = QString( cube->getAttribute( "CLUSTER ROOT CNODE ID" ).c_str() );
    if ( nrs.length() > 0 )
    {
        TreeItem* item    = 0;
        uint      cnodeID = nrs.toUInt();
        foreach( TreeItem * it, treeItemHash.values() )
        {
            uint itemCnode = it->getCubeObject()->get_id();
            if ( itemCnode == cnodeID )
            {
                item = it;
                break;
            }
        }
        setAsLoop( item );
    }

    return top;
}

QString
CallTree::getItemName( cube::Vertex* vertex )
{
    std::string name = ( static_cast<cube::Cnode* > ( vertex ) )->get_callee()->get_name();

    std::vector<std::pair<std::string, double> > num_params = ( static_cast<cube::Cnode* > ( vertex ) )->get_num_parameters();
    std::string                                  sep        = " (";
    std::string                                  end        = "";
    for ( unsigned i = 0; i < num_params.size(); i++ )
    {
        char buffer[ 100 ];
        sprintf( buffer, "%g", num_params[ i ].second );
        name += sep + num_params[ i ].first + "=" + std::string( buffer );
        sep   = ", ";
        end   = ")";
    }
    std::vector<std::pair<std::string, std::string> > str_params = ( static_cast<cube::Cnode* > ( vertex ) )->get_str_parameters();
    for ( unsigned i = 0; i < str_params.size(); i++ )
    {
        name += sep + str_params[ i ].first + "=" + str_params[ i ].second;
        sep   = ", ";
        end   = ")";
    }
    name += end;

    return QString( name.c_str() );
}


// -----------------------------------------------------------------
// ---------------- iteration handling --------------------------
// -----------------------------------------------------------------



/** The given item is set as a loop. This allows to hide its iterations.
 */
void
CallTree::setAsLoop( TreeItem* item )
{
    assert( item );
    if ( item == aggregatedLoopRoot || item == loopRoot )
    {
        return;
    }

    if ( item->children.size() == 0 )
    {
        return;
    }

    unsetLoop(); // unset previously set loop
    loopRoot           = item;
    aggregatedLoopRoot = mergeIterations( loopRoot );

    int iterations = aggregatedLoopRoot->iterations.size();
    aggregatedLoopRoot->name.append( " [" + QString::number( iterations ) + tr( " iterations" ) + "]" );

    loopRoot->displayName.clear();
    loopRoot->invalidateLabel(); // update tree item label of new loop root
}

void
CallTree::unsetLoop()
{
    if ( loopRoot != 0 )
    {
        if ( iterationsAreHidden_ )
        {
            showIterations();
        }

        deleteMergedIterations( aggregatedLoopRoot );
        TreeItem* oldLoop = loopRoot;
        aggregatedLoopRoot = 0;
        loopRoot           = 0;

        oldLoop->displayName.clear();
        oldLoop->invalidateLabel(); // update tree item label after loopRoot has been changed
    }
}


/**
   Hides the iteration of the currenly defined loop.
 */
void
CallTree::hideIterations()
{
    if ( !loopRoot  || iterationsAreHidden_ )
    {
        return;
    }
    iterationsAreHidden_ = true;

    // replace loop structure with aggregated structure
    treeModel->replaceSubtree( loopRoot, aggregatedLoopRoot );

    aggregatedLoopRoot->select();
    aggregatedLoopRoot->setExpanded( true );

    updateTreeItemList();

    aggregatedLoopRoot->displayName.clear();
    aggregatedLoopRoot->invalidateLabel();
}

void
CallTree::showIterations()
{
    if ( !loopRoot  || !iterationsAreHidden_ )
    {
        return;
    }
    iterationsAreHidden_ = false;

    // replace aggregated loop structure with iterations
    treeModel->replaceSubtree( aggregatedLoopRoot, loopRoot );

    loopRoot->select();
    loopRoot->setExpanded( true );

    updateTreeItemList();

    loopRoot->displayName.clear();
    loopRoot->invalidateLabel();
}

bool
CallTree::iterationsAreHidden()
{
    return iterationsAreHidden_;
}

TreeItem*
CallTree::getLoopRootItem()
{
    return loopRoot;
}

AggregatedTreeItem*
CallTree::getAggregatedRootItem()
{
    return aggregatedLoopRoot;
}

/**
   Treats the current item as root of a loop and merges its children. References from the merged iterations
   to the single iterations and vice versa are inserted.
 */
AggregatedTreeItem*
CallTree::mergeIterations( TreeItem* loopRoot )
{
    AggregatedTreeItem* aggregatedRoot = new AggregatedTreeItem( this, loopRoot->getDepth(), loopRoot );
    aggregatedRoot->cubeObject = loopRoot->cubeObject;
    aggregatedRoot->parentItem = loopRoot->parentItem;

    foreach( TreeItem * iteration, loopRoot->getChildren() )
    {
        aggregatedRoot->iterations.append( static_cast<Cnode*> ( iteration->getCubeObject() ) );
    }

    mergeIterations( aggregatedRoot, loopRoot->getChildren() );

    return aggregatedRoot;
}


/**
 * @brief mergeIterations recursively merges iteration items with the same name into
 * new items and appends them to newParent.
 * Only items of the same tree depth which have the same name are merged.
 */
void
CallTree::mergeIterations( TreeItem* newParent, const QList<TreeItem*>& iterations )
{
    QSet<QString>                       allChildren; // all distinct children names of the current tree level
    QHash<QString, AggregatedTreeItem*> hash;        // child name -> aggregated tree item

    // find all children of iterations, eleminate those with same name and append a copy to newparent
    foreach( TreeItem * iteration, iterations )
    {
        if ( iteration )
        {
            foreach( TreeItem * child, iteration->getChildren() )
            {
                allChildren.insert( child->name );
                if ( !hash.contains( child->name ) ) // create pseudo-item which contains iterations
                {
                    AggregatedTreeItem* newChild = new AggregatedTreeItem( this, newParent->getDepth() + 1, child );
                    newParent->addChild( newChild );
                    hash.insert( child->name, newChild );
                }
                //AggregatedTreeItem* aggregated = hash.value( child->name );
                //aggregated->iterations.append( static_cast<cube::Cnode*> ( child->cubeObject ) );
            }
        }
    }

    // required if functions are not called in all iterations -> append null cnode for non existing iterations
    // (optimize: inner loop not necessary if function is called in all iterations)
    foreach( TreeItem * iteration, iterations )
    {
        foreach( QString childName, allChildren.values() )
        {
            AggregatedTreeItem* aggregated = hash.value( childName );
            cube::Cnode*        cnode      = 0;
            if ( iteration )
            {
                foreach( TreeItem * currentIterChild, iteration->getChildren() ) // seach child in current iteration
                {
                    if ( childName == currentIterChild->name )
                    {
                        cnode = static_cast<cube::Cnode*> ( currentIterChild->cubeObject );
                        break;
                    }
                }
                if ( cnode )
                {
                    aggregated->iterations.append( cnode );
                }
            }
        }
    }

    // if newly added children of newParent have children, call this method recursively
    foreach( TreeItem * item, newParent->getChildren() )
    {
        QString childName = item->name;

        int              childCount = 0;
        QList<TreeItem*> listOfChildren;
        foreach( TreeItem * iteration, iterations )
        {
            TreeItem* nextChild = 0;
            if ( iteration )
            {
                foreach( TreeItem * child, iteration->getChildren() )
                {
                    if ( child->name == childName )
                    {
                        nextChild = child;
                        childCount++;
                        break;
                    }
                }
            }
            listOfChildren.push_back( nextChild );
        }

        if ( childCount > 0 )
        {
            mergeIterations( item, listOfChildren );
        }
    }
}

/**
   delete the merged iterations and all its children recursively
 */
void
CallTree::deleteMergedIterations( TreeItem* aggregated )
{
    foreach( TreeItem * item, aggregated->getChildren() )             // for all aggregated items
    {
        deleteMergedIterations( item );
        delete item;
    }
}


/**
 * updates the vector Tree::treeItems by exchanging loop iterations with aggregated iterations and vice versa
 */
void
CallTree::updateTreeItemList()
{
    if ( (  iterationsAreHidden_ && treeItems.contains( aggregatedLoopRoot ) ) ||
         ( !iterationsAreHidden_ && treeItems.contains( loopRoot ) ) )
    {
        return;             // already up to date
    }

    QList<TreeItem*> items;
    items.append( this->top->getChildren() );

    treeItems.clear();

    // recursively add current tree elements to treeItems list
    while ( !items.empty() )
    {
        TreeItem* item = items.takeFirst();
        treeItems.append( item );
        if ( item->getChildren().size() > 0 )
        {
            items.append( item->getChildren() );
        }
    }
}

// -----------------------------------------------------------------
// ---------------- end iterations ------ --------------------------
// -----------------------------------------------------------------



// -----------------------------------------------------------------
// ---------------- calculation functions --------------------------
// -----------------------------------------------------------------

/**
 * @brief getSelectedFromHiddenIterations returns a list of items of the hidden iterations which correspond
 * to the selected merged item
 * @param iterations a list of hidden iterations, starting with tree level after root
 * @param selected path to the selected item, contains the names of each ancestor of the selected item,
 * starting with the level after root
 * @return list with the selected item for each iteration where an item with the same name at the same tree
 * level is found
 */
QList<TreeItem*>
getSelectedFromHiddenIterations( const QList<TreeItem*>& iterations, QStringList& selected )
{
    QList<TreeItem*> nextLevel;
    QList<TreeItem*> ret;

    QString selectedName = selected.takeFirst();
    foreach( TreeItem * iter, iterations )
    {
        foreach( TreeItem * child, iter->getChildren() )
        {
            if ( child->getName() == selectedName )             // found selected item in different iteration
            {
                nextLevel.push_back( child );
            }
        }
    }
    if ( !selected.empty() )
    {
        ret = getSelectedFromHiddenIterations( nextLevel, selected );
    }
    else
    {
        ret = nextLevel;
    }
    return ret;
}

QVector<Task*>
CallTree::setBasicValues( const QList<Tree*>& leftTrees,
                          const QList<Tree*>& rightTrees, const QList<TreeItem*>& itemsToCalculate )
{
    QVector<Task*> workerData;
    /*
     * No update of tree is necessary if the call tree is to the left of the metric tree
     */
    if ( ( leftTrees.size() == 0 ) || ( ( leftTrees.size() == 1 ) && ( leftTrees[ 0 ]->getTreeType() != METRICTREE ) ) )
    {
        return workerData;
    }

    if ( itemsToCalculate.size() == 0 ) // all visible items are already up to date
    {
        return workerData;
    }

    // Set up tree pointers
    Tree* metricTree        = 0;
    Tree* systemTree        = 0;
    bool  rightOfSystemTree = false;

    // Metric tree is definitely in the left trees, as the initial if-statement at the
    // beginning of the call already filtered out the other cases.
    foreach( Tree * item, leftTrees )
    {
        if ( item->getTreeType() == METRICTREE )
        {
            metricTree = item;
        }
        else if ( item->getTreeType() == SYSTEMTREE )
        {
            systemTree        = item;
            rightOfSystemTree = true;
        }
    }
    assert( metricTree );

    // if system tree was not in the left trees, search in the right trees
    if ( !systemTree )
    {
        foreach( Tree * item, rightTrees )
        {
            if ( item->getTreeType() == SYSTEMTREE )
            {
                systemTree = item;
            }
        }
    }
    assert( systemTree );

    // Set up selection for metric tree
    const QList<TreeItem*>& selected_metrics = metricTree->getSelectionList();
    list_of_metrics         metric_selection;

    foreach( TreeItem * item, selected_metrics )
    {
        metric_pair mp;
        mp.first  = static_cast<cube::Metric*> ( item->getCubeObject() );
        mp.second = ( item->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
        if ( item->getChildren().size() == 0 )
        {
            mp.second = CUBE_CALCULATE_INCLUSIVE;
        }
        metric_selection.push_back( mp );
    }

    // Set up selection for system tree, leave empty if system tree is right to call tree
    list_of_sysresources sysres_selection;
    if ( rightOfSystemTree )  // call tree is on the right
    {
        QList< TreeItem* > selected_systemnodes = systemTree->getSelectionList();
        foreach( TreeItem * item, selected_systemnodes )
        {
            sysres_pair sp;
            sp.first  = static_cast<cube::Sysres*> ( item->getCubeObject() );
            sp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sysres_selection.push_back( sp );
        }
    }
    else
    {
        sysres_selection = CubeProxy::ALL_SYSTEMNODES;
    }

    // get values for full tree
    int depth = 0;              // 0 = first level; todo optimize
    foreach( TreeItem * child, itemsToCalculate )
    {
        cube::Cnode* cnode = 0;
        if ( child->isAggregatedRootItem() ) // special case aggregated loops: calculate complete subtree
        {
            // calculate all non aggregated loop items
            cnode = static_cast<cube::Cnode*> ( loopRoot->getCubeObject() );
            depth = getDepth() - child->getDepth(); // set depth to calculate complete original loop subtree
        }
        else if ( child->isAggregatedLoopItem() )
        {
            continue; // already calculated with parent loop root
        }
        else
        {
            cnode = static_cast<cube::Cnode*> ( child->getCubeObject() );
        }

        CallTreeData* data = new CallTreeData( child,
                                               metric_selection,
                                               sysres_selection,
                                               cnode,
                                               depth );
        workerData.append( data );
    }
    return workerData;
}


/** used to calculate the data of root_cnode in a thread */
void
CallTreeData::calculate()
{
    vector< Value* > inclusive_values;
    vector< Value* > exclusive_values;
    cube::IdIndexMap indexMap;
    CallTree*        tree = static_cast<CallTree*> ( item->getTree() );
    cube::CubeProxy* cube = tree->getCube();

    cube->getCallpathSubtreeValues( metric_selection,
                                    sysres_selection,
                                    *root_cnode,
                                    depth,
                                    indexMap,
                                    &inclusive_values,
                                    &exclusive_values );
    tree->assignValues( item, inclusive_values, exclusive_values, item->getDepth() + depth, &indexMap );
}


void
CallTree::assignValues( TreeItem*               item,
                        const vector< Value* >& inclusive_values,
                        const vector< Value* >& exclusive_values,
                        int                     depth,
                        cube::IdIndexMap*       indexMap
                        )
{
    assert( item );

    if ( item->getDepth() > depth )
    {
        return;
    }

    if ( item->getCubeObject() )
    {
        Tree::assignValues( item, inclusive_values, exclusive_values, depth, indexMap );
    }
    else // iterations
    {
        Value* inclusiveValue = selectValueOnType( inclusive_values.front() );
        Value* exclusiveValue = selectValueOnType( exclusive_values.front() );

        foreach( Cnode * contributingCnode, static_cast<AggregatedTreeItem*>( item )->getIterations() )
        {
            if ( contributingCnode ) // function has been called in current iteration
            {
                int id = contributingCnode->get_id();
                if ( indexMap )
                {
                    id = indexMap->getIndex( id );
                }
                *inclusiveValue += inclusive_values[ id ];
                *exclusiveValue += exclusive_values[ id ];
            }
        }

        item->setValues( inclusiveValue, exclusiveValue );

        foreach( TreeItem * child, item->getChildren() )
        {
            assignValues( child, inclusive_values, exclusive_values, depth, indexMap );
        }
    }

    // add exclusive values of the hidden child items the the aggregated loop root
    if ( item->isAggregatedRootItem() )
    {
        Value* exclusiveValue = item->ownValue_adv;                   // value of the aggregated loop root
        foreach( TreeItem * child, getLoopRootItem()->getChildren() ) // all hidden children
        {
            size_t id = child->getCubeObject()->get_id();
            if ( indexMap )
            {
                id = indexMap->getIndex( id );
            }
            *exclusiveValue += exclusive_values[ id ];
        }
    }
}

// -----------------------------------------------------------------
// ---------------- end calculation functions ----------------------
// -----------------------------------------------------------------


TreeItem*
CallTree::getTreeItem( uint32_t cnodeId )
{
    foreach( TreeItem * item, treeItems )
    {
        cube::Cnode* cnode = static_cast<cube::Cnode*> ( item->getCubeObject() );
        if ( cnode->get_id() == cnodeId )
        {
            return item;
        }
    }
    return 0;
}

/** sets current item as call tree root item
 */
void
CallTree::setItemAsRoot( TreeItem* newRoot )
{
    lastSelection = newRoot;
    selectionList.clear();
    selectionList.append( lastSelection );
    updateHidenItemsInSelectionList();

    // set item properties according to root position
    cube->rerootCnode( *static_cast<cube::Cnode* > ( newRoot->getCubeObject() ) );
    newRoot->setDepth( 0 );

    treeModel->setAsRoot( newRoot );

    // invalidate all trees left of the call tree
    TabManager*  tab = Globals::getTabManager();
    QList<Tree*> left, right;
    tab->getNeighborTrees( left, right, this );
    foreach( Tree * tree, left )
    {
        tree->invalidateItems();
    }
    tab->reinit(); // recalculate trees
}

/** removes the given item from the tree
 */
void
CallTree::pruneItem( TreeItem* item )
{
    cube::Cnode& cnode =  *static_cast<cube::Cnode* > ( item->getCubeObject() );
    treeModel->removeSubtree( item );
    cube->pruneCnode( cnode );
}

/** removes the children of the item from the tree
 */
void
CallTree::setItemAsLeaf( TreeItem* item )
{
    cube::Cnode& cnode =  *static_cast<cube::Cnode* > ( item->getCubeObject() );
    cube->setCnodeAsLeaf( cnode );
    treeModel->removeChildren( item );
}
