/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2020                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup CUBE_gui.network
 *  @brief   Definition of the class cube::QtStreamSocket.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "CubeQtStreamSocket.h"

#include <QApplication>
#include <QEventLoop>
#include <QMetaType>
#include <QTcpServer>
#include <QTcpSocket>
#include <QThread>
#include <QTimer>
#include <QtGlobal>

#include "CubeError.h"

using namespace std;
using namespace cube;

#if QT_VERSION < 0x050000 // already implemented in Qt5
Q_DECLARE_METATYPE( QAbstractSocket::SocketError );
Q_DECLARE_METATYPE( QAbstractSocket::SocketState );
#endif

#define SETUP_EVENT_LOOP_FOR( object, signal )                          \
    QTimer timer;                                                     \
    timer.setSingleShot( true );                                        \
    QEventLoop loop;                                                  \
    loop.connect( &timer, SIGNAL( timeout() ), SLOT( quit() ) );            \
    loop.connect( object, SIGNAL( signal ), SLOT( quit() ) );               \
    loop.connect( object, SIGNAL( error( QAbstractSocket::SocketError ) ), \
                  SLOT( quit() ) );

#define PROCESS_EVENT_LOOP() \
    loop.exec();

#define PROCESS_EVENT_LOOP_WITH_TIMEOUT( timeout ) \
    timer.start( timeout );                        \
    loop.exec();

#define ASSERT_STATE_OF( socket, socket_state )                     \
    if ( socket->state() != socket_state )                          \
    {                                                             \
        throw UnrecoverableNetworkError( QObject::tr( "Connection timed out." ).toUtf8().data() ); \
    }

namespace
{
// default socket timeout of 1 second
int DEFAULT_TIMEOUT = 10000;
}

SocketPtr
QtStreamSocket::create()
{
    qRegisterMetaType< QAbstractSocket::SocketError >(
        "QAbstractSocket::SocketError" );
    qRegisterMetaType< QAbstractSocket::SocketState >(
        "QAbstractSocket::SocketState" );

    return SocketPtr( new QtStreamSocket() );
}


void
QtStreamSocket::connect( const std::string& address,
                         int                port )
{
    if ( mSocket )
    {
        if ( mSocket->isOpen() )
        {
            disconnect();
        }
    }
    else
    {
        mSocket = new QTcpSocket();
        mSocket->moveToThread( QApplication::instance()->thread() );
    }

    if ( !mSocket )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Cannot connect on invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mSocket, connected() );

    mSocket->connectToHost( QString( address.c_str() ), port );

    PROCESS_EVENT_LOOP_WITH_TIMEOUT( ::DEFAULT_TIMEOUT );

    ASSERT_STATE_OF( mSocket, QAbstractSocket::ConnectedState );

    if ( !mSocket->isValid() )
    {
        throw UnrecoverableNetworkError( mSocket->errorString().toStdString() );
    }
}


void
QtStreamSocket::disconnect()
{
    if ( !mSocket || !mSocket->isValid() || !mSocket->isOpen() )
    {
        throw UnrecoverableNetworkError(
                  QObject::tr( "Attempting to disconnect invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mSocket, disconnected() );

    mSocket->disconnectFromHost();

    PROCESS_EVENT_LOOP_WITH_TIMEOUT( 50 );
}


void
QtStreamSocket::listen( int port )
{
    if ( mServer )
    {
        shutdown();
    }
    else
    {
        mServer = new QTcpServer();
        mServer->moveToThread( QApplication::instance()->thread() );
    }

    if ( !mServer )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Cannot listen on invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mServer, stateChanged() );

    mServer->listen( QHostAddress::Any, port );

    PROCESS_EVENT_LOOP_WITH_TIMEOUT( ::DEFAULT_TIMEOUT );

    if ( !mServer->isListening() )
    {
        throw UnrecoverableNetworkError( mServer->errorString().toStdString() );
    }
}


void
QtStreamSocket::accept()
{
    if ( !mServer )
    {
        throw UnrecoverableNetworkError(
                  QObject::tr( "Cannot accept connections on invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mSocket, connected() );

    mSocket = mServer->nextPendingConnection();

    PROCESS_EVENT_LOOP_WITH_TIMEOUT( ::DEFAULT_TIMEOUT );

    if ( !mSocket->isOpen() )
    {
        throw UnrecoverableNetworkError( mSocket->errorString().toStdString() );
    }
}


void
QtStreamSocket::shutdown()
{
    if ( !mServer )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Cannot shutdown invalid socket." ).toUtf8().data() );
    }
    mServer->close();
}


size_t
QtStreamSocket::send( const void* buffer,
                      size_t      num_bytes )
{
    if ( !mSocket || !mSocket->isValid() || !mSocket->isOpen() )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Unable to write to invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mSocket, bytesWritten( qint64 ) );

    qint64 bytesSent  = 0;
    qint64 totalBytes = num_bytes;

    while ( bytesSent < totalBytes )
    {
        qint64 res =
            mSocket->write( &static_cast< const char* >( buffer )[ bytesSent ],
                            totalBytes - bytesSent );
        mSocket->flush();

        if ( res < 0 )
        {
            throw UnrecoverableNetworkError( mSocket->errorString().toStdString() );
        }
        else
        {
            bytesSent += res;
        }

        while ( mSocket->bytesToWrite() > 0 )
        {
            PROCESS_EVENT_LOOP_WITH_TIMEOUT( 100 );

            ASSERT_STATE_OF( mSocket, QAbstractSocket::ConnectedState );

            if ( !mSocket->isWritable() )
            {
                throw UnrecoverableNetworkError(
                          mSocket->errorString().toStdString() );
            }
        }
    }

    return bytesSent;
}


size_t
QtStreamSocket::receive( void*  buffer,
                         size_t num_bytes )
{
    if ( !mSocket || !mSocket->isValid() || !mSocket->isOpen() )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Unable to read from invalid socket." ).toUtf8().data() );
    }

    SETUP_EVENT_LOOP_FOR( mSocket, readyRead() );

    qint64 bytesRead  = 0;
    qint64 totalBytes = num_bytes;

    while ( bytesRead < totalBytes )
    {
        if ( mSocket->bytesAvailable() > 0 )
        {
            qint64 res =
                mSocket->read( &static_cast< char* >( buffer )[ bytesRead ],
                               qMin( totalBytes - bytesRead,
                                     mSocket->bytesAvailable() ) );
            if ( res < 0 )
            {
                throw UnrecoverableNetworkError(
                          mSocket->errorString().toStdString() );
            }
            else
            {
                bytesRead += res;
            }
        }
        else
        {
            PROCESS_EVENT_LOOP_WITH_TIMEOUT( 100 );

            ASSERT_STATE_OF( mSocket, QAbstractSocket::ConnectedState );

            if ( !mSocket->isReadable() )
            {
                throw UnrecoverableNetworkError(
                          mSocket->errorString().toStdString() );
            }
        }
    }

    return bytesRead;
}


bool
QtStreamSocket::isConnected()
{
    if ( !mSocket )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Invalid socket." ).toUtf8().data() );
    }

    return mSocket->isOpen();
}


bool
QtStreamSocket::isListening()
{
    if ( !mServer )
    {
        throw UnrecoverableNetworkError( QObject::tr( "Invalid socket." ).toUtf8().data() );
    }

    return mServer->isListening();
}


std::string
QtStreamSocket::getHostname()
{
    return QObject::tr( "Unknown host" ).toUtf8().data();
}


std::string
QtStreamSocket::getInfoString()
{
    return "Qt";
}


QtStreamSocket::QtStreamSocket()
    : mSocket( NULL ),
    mServer( NULL )
{
}


QtStreamSocket::~QtStreamSocket()
{
    if ( mSocket )
    {
        if ( isConnected() )
        {
            disconnect();
        }
        delete mSocket;
    }
    if ( mServer )
    {
        if ( isListening() )
        {
            shutdown();
        }
        delete mServer;
    }
}
