/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>
#include <iostream>
#include "AlgMgmtTest.h"

using namespace advisor;

std::string AlgMgmtTest::comment_poor = QObject::tr(
    "Application spends TOO MUCH amount of its run time doing "
    "management actions such as MPI setup, synchronization between processes. "
    "<br>"
    "It is worth to try to reduce this in order to use compute time more effectively." ).toStdString();

std::string AlgMgmtTest::comment_significant = QObject::tr(
    "Application spends SIGNIFICANT amount of its run time doing "
    "management actions such as MPI setup, synchronization between processes."
    "<br>"
    "It is worth to try to reduce this in order to use compute time more effectively."  ).toStdString();
std::string AlgMgmtTest::comment_some = QObject::tr(
    "Application spends SOME amount of its run time doing "
    "management actions such as MPI setup, synchronization between processes."
    "<br>"
    "It is worth to try to reduce this in order to use compute time more effectively."  ).toStdString();
std::string AlgMgmtTest::comment_notime = QObject::tr(
    "Application spends nearly no time in algorithm management which allows it to <br>"
    "use HPC resources more effectively." ).toStdString();





AlgMgmtTest::AlgMgmtTest( cube::CubeProxy* cube ) : PerformanceTest( cube )
{
    setName( tr( "Algorithm Management" ).toUtf8().data() );
    setWeight( 0.2 );   // need to be adjusted
    setValue( 0 );
    management = cube->getMetric( "mgmt" );
    sync       = cube->getMetric( "sync" );
    execution  = cube->getMetric( "execution" );

    if ( management == NULL ||
         sync == NULL ||
         execution == NULL
         )
    {
        return;
    }

    cube::list_of_metrics      lmetrics;
    cube::list_of_sysresources lsysres = getRootsOfSystemTree();
    cube::metric_pair          metric;

    cube::value_container mgm_inclusive_values;
    cube::value_container mgm_exclusive_values;
    cube::value_container sync_inclusive_values;
    cube::value_container sync_exclusive_values;
    cube::value_container exec_inclusive_values;
    cube::value_container exec_exclusive_values;


    metric.first  = management;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
    cube->getCallpathTreeValues( lmetrics,
                                 lsysres,
                                 mgm_inclusive_values,
                                 mgm_exclusive_values );
    lmetrics.clear();
    metric.first  = sync;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
    cube->getCallpathTreeValues( lmetrics,
                                 lsysres,
                                 sync_inclusive_values,
                                 sync_exclusive_values );
    lmetrics.clear();
    metric.first  = execution;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
    cube->getCallpathTreeValues( lmetrics,
                                 lsysres,
                                 exec_inclusive_values,
                                 exec_exclusive_values );
    for ( size_t i = 0; i < mgm_inclusive_values.size(); ++i )
    {
        double mgm_ivalue  = ( mgm_inclusive_values[ i ] )->getDouble();
        double sync_ivalue = ( sync_inclusive_values[ i ] )->getDouble();
        double exec_ivalue = ( exec_inclusive_values[ i ] )->getDouble();

        double mgm_evalue  = ( mgm_exclusive_values[ i ] )->getDouble();
        double sync_evalue = ( sync_exclusive_values[ i ] )->getDouble();
        double exec_evalue = ( exec_exclusive_values[ i ] )->getDouble();

        cube::Value* i_value = ( exec_inclusive_values[ i ] )->clone();
        i_value->
        operator=( 1 - ( mgm_ivalue + sync_ivalue ) / exec_ivalue  );

        inclusive_values.push_back( i_value );
        cube::Value* e_value = ( exec_exclusive_values[ i ] )->clone();
        e_value->
        operator=( 1 - ( mgm_evalue + sync_evalue ) / exec_evalue  );

        exclusive_values.push_back( e_value );
    }
}



void
AlgMgmtTest::applyCnode( const cube::Cnode* cnode, const cube::CalculationFlavour cnf  )
{
    double alg_mgm_value = ( cnf == cube::CUBE_CALCULATE_INCLUSIVE ) ?
                           inclusive_values[ cnode->get_id() ]->getDouble() :
                           exclusive_values[ cnode->get_id() ]->getDouble();
    setValue( alg_mgm_value );
}

void
AlgMgmtTest::applyCnode( const cube::list_of_cnodes& cnodes )
{
    if ( cnodes.size() == 1 )
    {
        applyCnode( cnodes[ 0 ].first, cnodes[ 0 ].second );
    }
    else
    {
        cube::value_container mgm_inclusive_values;
        cube::value_container mgm_exclusive_values;
        cube::value_container sync_inclusive_values;
        cube::value_container sync_exclusive_values;
        cube::value_container exec_inclusive_values;
        cube::value_container exec_exclusive_values;

        cube::list_of_sysresources lsysres = getRootsOfSystemTree();
        cube::IdIndexMap           metric_id_indices;
        cube->getMetricSubtreeValues( cnodes,
                                      lsysres,
                                      *management,
                                      1,
                                      metric_id_indices,
                                      &mgm_inclusive_values,
                                      &mgm_exclusive_values );

        cube->getMetricSubtreeValues( cnodes,
                                      lsysres,
                                      *sync,
                                      1,
                                      metric_id_indices,
                                      &sync_inclusive_values,
                                      &sync_exclusive_values );

        cube->getMetricSubtreeValues( cnodes,
                                      lsysres,
                                      *execution,
                                      1,
                                      metric_id_indices,
                                      &exec_inclusive_values,
                                      &exec_exclusive_values );

        double alg_mgm_value = 1 - ( ( mgm_inclusive_values[ 0 ]->getDouble() ) + ( sync_inclusive_values[ 0 ]->getDouble() ) ) / ( exec_inclusive_values[ 0 ]->getDouble() );
        setValue( alg_mgm_value );
    }
}


const std::string&
AlgMgmtTest::getCommentText() const
{
    int range = value() * 100 / 33;
    switch ( range )
    {
        case 0:
            return comment_poor;
        case 1:
            return comment_significant;
        case 2:
            return comment_some;
        case 3:
            return comment_notime;
    }
    return no_comment;
}


// ------ overview tests ---------

bool
AlgMgmtTest::isActive() const
{
    return ( management != NULL ) && ( sync != NULL ) && ( execution != NULL );
};

bool
AlgMgmtTest::isIssue() const
{
    return false;
};
